// Some library wide, skeleton aware, utilities for constructing and traversing call trees
// None of these are visible from the library API

int DEBUGCT=0; // Set to 1 for debug informations

// For debugging, defined later
void DisplayCallTree (callnode_t *t);

// Auxilliary tree building function, see later
void AddtoCallTree (callnode_t **tree, int *callstack, int ncalls);

// Build the call tree (complete copy locally to each process) for the current
// BUFF rooted skeleton nest
// Root it at FrStack[Top]->mycalltree
void BuildCallTree (void) {
  int i, j;
  int filler;
  int actcount = 0;
  int mytotal;      // how much info do I have? (in ints)
  int *myinfo;
  int *infosizes;
  int infostart;
  int alltotalsize;
  int *allrawinfo;
  int *recvcounts, *displs;
  int ROOT;
  int rootcommsize;
  MPI_Comm rootcomm;

  if (DEBUGCT) printf("Entering BuildCallTree - %d - nestlevel %d \n", FrStack[0]->myrank, FrStack[Top]->nestlevel);

  ROOT = Top - FrStack[Top]->nestlevel + 1;  

  mytotal = (FrStack[Top]->nestlevel)*5 + 2;     // see structure of call info for explanation
  myinfo  = (int *) malloc (sizeof(int)*mytotal);

  // Fill in the info buffer  
  myinfo[0]         = FrStack[Top]->nestlevel;
  myinfo[mytotal-1] = FrStack[Top]->myrank;

  filler = 1;
  for (i=FrStack[Top]->nestlevel-1; i>=0 ; i--) {
    myinfo[filler++]   = FrStack[Top-i]->skel;
    switch (FrStack[Top-i]->skel) {
    case Pskel: 
      myinfo[filler++] = FrStack[Top-i]->parentcommrank;
      myinfo[filler++] = FrStack[Top-i]->context.p.ns;
      myinfo[filler++] = FrStack[Top-i]->mycol;
      myinfo[filler++] = FrStack[Top-i]->mycommsize;
      break;
    case Dskel: 
      myinfo[filler++] = FrStack[Top-i]->parentcommrank;
      myinfo[filler++] = FrStack[Top-i]->context.d.nw;
      myinfo[filler++] = FrStack[Top-i]->mycol;
      myinfo[filler++] = FrStack[Top-i]->mycommsize;
      break;
    default: 
      fprintf(stderr,"BuildCallTree for an unimplemented skeleton\n");
      exit(0);
      break;
    }
  }

  // Now all-all exchange info buffers within the BUF mode root comm
  rootcomm     = FrStack[ROOT]->parentcomm;
  rootcommsize = FrStack[ROOT]->parentcommsize;

  // Allocate space for one count per process in the root group
  infosizes = (int *) malloc (sizeof(int)*rootcommsize);

  MPI_Allgather(&mytotal,1,MPI_INT,infosizes,1,MPI_INT,rootcomm);

  // Build parameters for the real exchange
  recvcounts = (int *) malloc (sizeof(int)*rootcommsize);
  displs     = (int *) malloc (sizeof(int)*rootcommsize);
  displs[0]  = 0;
  for (i=0; i<rootcommsize-1; i++) {
    recvcounts[i] = infosizes[i];
    displs[i+1]   = displs[i] + infosizes[i];
  }
  recvcounts[rootcommsize-1] = infosizes[rootcommsize-1];
  alltotalsize = displs[rootcommsize-1]+infosizes[rootcommsize-1];
  allrawinfo = (int *) malloc (sizeof(int)*alltotalsize);

  MPI_Allgatherv(myinfo,mytotal,MPI_INT,allrawinfo,recvcounts,displs,MPI_INT,rootcomm);

  // Build the tree, adding info from procs one at a time from the allrawinfobuffer
  infostart = 0;
  for (i=0; i<rootcommsize; i++) {
    AddtoCallTree (&FrStack[Top]->mycalltree, &allrawinfo[infostart], allrawinfo[infostart]);
    infostart += infosizes[i];
  }
  free(displs); free(recvcounts); free(allrawinfo); free(infosizes); free(myinfo);

  if (DEBUGCT) printf("Exiting BuildCallTree - %d\n", FrStack[0]->myrank);

}

// Add the single process call tree info for the last "ncalls" calls from callstack[...] 
// to the call tree pointed to by *tree. Note that callstack is always the complete
// callstack, into which we have to jump to find the appropriate frames
// 
// Precondition: ncalls >= 1
void AddtoCallTree (callnode_t **tree, int *callstack, int ncalls) {

  int i, j;
  int *thiscall = &callstack[1+5*(callstack[0]-ncalls)];  // start of this call frame

  // if this is the first time we have seen this call, build a node for it
  if (*tree == NULL) {
    *tree = (callnode_t *) malloc (sizeof(callnode_t));
    (*tree)->skel    = thiscall[0];

    // is this my own call history?
    (*tree)->comm    = (callstack[2]==(FrStack[Top-callstack[0]+1]->parentcommrank)) ? FrStack[Top-ncalls+1]->parentcomm : (MPI_Comm) NULL;

    (*tree)->Nacts   = thiscall[2];
    (*tree)->level   = callstack[0]-ncalls;
    (*tree)->actsz   = (int *)  calloc ((*tree)->Nacts, sizeof(int));
    (*tree)->members = (int **) calloc ((*tree)->Nacts, sizeof(int *));
    (*tree)->acts    = (callnode_t **) calloc ((*tree)->Nacts,sizeof(callnode_t *));
    (*tree)->Gcount  = 0; (*tree)->Tcount  = 0; (*tree)->Ecount  = 0;
  }

  // is this the first visitor from this activity? 
  if ((*tree)->actsz[thiscall[3]] == 0) {
    (*tree)->actsz[thiscall[3]] = thiscall[4];
    // now create the members array for this act only,
    // initialising to -1 so that we can fill in more later
    (*tree)->members[thiscall[3]] = (int *) malloc(sizeof(int)*thiscall[4]*((*tree)->level+1));
    // initialise
    for (i=0; i<thiscall[4]*((*tree)->level+1); i++) { 
      (*tree)->members[thiscall[3]][i] = -1;
    }
  }


  // now fill in this caller's id history
  j=0;
  while ((*tree)->members[thiscall[3]][j] >= 0) j++;         // find the place, *VERY INEFFICIENT* for big P
  for (i=0; i<(*tree)->level+1; i++) { 
    (*tree)->members[thiscall[3]][j+i] = callstack[1+5*i+1]; // ie copy from myid field of each frame
  }


  // work our way down the tree recursively
  if (ncalls > 1) {
    if (DEBUGCT) {
      printf("Recursive call to AddtoCallTree -- %d \n", ncalls);
      DisplayCallTree(*tree);
    }
    // *** there is a problem with this call in some cases of 3-level nesting
    // *** needs to be studied and fixed
    AddtoCallTree (&((*tree)->acts[thiscall[3]]), callstack, ncalls-1);
  }
  
  else {
    // if necessary, build the special leaf node
    if ((*tree)->acts[thiscall[3]] == NULL) {
      (*tree)->acts[thiscall[3]] = (callnode_t *) malloc (sizeof(callnode_t));
      (*tree)->acts[thiscall[3]]->skel = NOskel;
      // is the my own call history?
      (*tree)->acts[thiscall[3]]->comm = (callstack[2] == FrStack[Top-callstack[0]+1]->parentcommrank) ? 
                                          FrStack[Top-ncalls+1]->mycomm : (MPI_Comm) NULL;

      (*tree)->acts[thiscall[3]]->Nacts = 0;
      (*tree)->acts[thiscall[3]]->level = callstack[0];
      (*tree)->acts[thiscall[3]]->actsz = (int *) malloc (sizeof(int));
      (*tree)->acts[thiscall[3]]->actsz[0] = thiscall[4];
      (*tree)->acts[thiscall[3]]->members = (int **) malloc (sizeof(int *));

      // now add this caller's details
      (*tree)->acts[thiscall[3]]->members[0] = (int *) malloc (sizeof(int)*thiscall[4]*((*tree)->level+2));
      for (i=0; i<thiscall[4]*((*tree)->level+2); i++) { 
        (*tree)->acts[thiscall[3]]->members[0][i]= -1;
      }
    }
    
    j=0;
    while ((*tree)->acts[thiscall[3]]->members[0][j] >= 0) j++; 
    // find the place, *VERY INEFFICIENT* for big P
    
    for (i=0; i<(*tree)->level+1; i++) { 
      (*tree)->acts[thiscall[3]]->members[0][j+i] = callstack[1+5*i+1];
      // ie copy from myid field of each frame
    }
    (*tree)->acts[thiscall[3]]->members[0][j+callstack[0]] = callstack[1+5*callstack[0]]; // special case
  }
}


void indent (int gap) {
  int i;
  for (i=0; i<gap*10; i++) {
    printf(" ");
  }
}

void DisplayCallTree (callnode_t *t) {
  int i,j,k;

  indent(t->level);
  printf("==========================\n");
  switch (t->skel) {
  case NOskel: 
    indent(t->level);
    printf("LEAF\n");
    break;
  case Pskel: 
    indent(t->level);
    printf("PIPELINE\n");
    break;
  case Dskel: 
    indent(t->level);
    printf("DEAL\n");
    break;
  default: 
    indent(t->level);
    printf("UNKNOWN SKEL TAG\n");
    break;
  }
  if (t->comm) {
    indent(t->level);
    printf("local comm\n");
  } else {
    indent(t->level);
    printf("remote comm\n");
  }
  indent(t->level);
  printf("%d sub-activities\n",t->Nacts);
  indent(t->level);
  printf("level %d\n",t->level);

  if (t->skel == NOskel) {
  indent(t->level);
  printf("actsize : ");
      printf("%d ",t->actsz[0]);
  } else {
    indent(t->level);
    printf("actsizes : ");
    for (i=0; i<t->Nacts; i++) {
      printf("%d ",t->actsz[i]);
    }
  }
  printf("\n");

  if (t->skel != NOskel) {
    for (i=0; i<t->Nacts; i++) {
      indent(t->level);
      printf("Act %d : ",i);
      printf(" || ");
      for (j=0; j<t->actsz[i]; j++) {
        for (k=0; k<t->level+1; k++) {
          printf("%d ",t->members[i][j*(t->level+1) + k]);
        }
        printf(" || ");
      }
      printf("\n");
    }
    indent(t->level);
    printf("==========================\n");
    printf("\n"); 
    for (i=0; i<t->Nacts; i++) {
      DisplayCallTree(t->acts[i]);
    } 
  } else {
    indent(t->level);
    printf("Leaf members : ");
    printf(" || ");
    for (j=0; j<t->actsz[0]; j++) {
      for (k=0; k<t->level+1; k++) {
        printf("%d ",t->members[0][j*(t->level+1) + k]);
      }
      printf("|| ");
    }
    printf("\n");
    indent(t->level);
    printf("==========================\n");
    printf("\n");
  }
}


// Traverse the current call tree to find and return the parent of the leaf group for this process
callnode_t *GetMyParentGroup(void) {
  int rootindex;
  callnode_t * myparentnode;

  // Find the stack frame for the root of this nest
  rootindex = Top;
  while (FrStack[rootindex]->dmode != BUF) rootindex--;
 
  // use info on FrStack to traverse the call stack to my group
  myparentnode = FrStack[Top]->mycalltree;
  while (rootindex != Top) {
    myparentnode = myparentnode->acts[FrStack[rootindex]->mycol];
    rootindex++;
  }
  return myparentnode;
}

// Traverse the current call tree to find and return the leaf group for this process
callnode_t *GetMyGroup(void) {
  int rootindex;
  callnode_t * mynode;

  // Find the stack frame for the root of this nest
  rootindex = Top;
  while (FrStack[rootindex]->dmode != BUF) rootindex--;
 
  // use info on FrStack to traverse the call stack to my group
  mynode = FrStack[Top]->mycalltree;
  while (rootindex != Top) {
    mynode = mynode->acts[FrStack[rootindex]->mycol];
    rootindex++;
  }
  return mynode->acts[FrStack[rootindex]->mycol];
}


// Function which finds the fringe leaf activity group within CT which will produce the
// Nth output (ie "Give") from the CT top level skeleton.
callnode_t *FringeGiveGroup (callnode_t *CT, int N) {
  switch (CT->skel) {
  case NOskel: 
    return CT;
    break;
  case Pskel:
    // N is irrelevant for a pipeline
    return FringeGiveGroup (CT->acts[CT->Nacts - 1], N);
    break;
  case Dskel:
    // N is relevant for a deal: access the node N modulo nw
    return FringeGiveGroup (CT->acts[N % CT->Nacts], N);
    break;
 default:
    printf("FringeGiveGroup looking at an unimplemented skeleton\n");
    exit(0);
  }
}


// Function which finds the fringe leaf activity group within CT which will consume the
// Nth input (ie "Take") from the CT top level skeleton.
callnode_t *FringeTakeGroup (callnode_t *CT, int N) {
  switch (CT->skel) {
  case NOskel: 
    return CT;
    break;
  case Pskel:
    // N is irrelevant for a pipeline
    return FringeTakeGroup (CT->acts[0], N);
    break;
  case Dskel:
    // N is relevant for a deal: access the node N modulo nw
    return FringeTakeGroup (CT->acts[N%CT->Nacts], N);
  default:
    printf("FringeTakeGroup looking at an unimplemented skeleton\n");
    exit(0);
  }
}


// Find the group which holds the input buffer for CT
callnode_t *InBuffGroup (callnode_t *CT) {
  switch (CT->skel) {
  case NOskel: 
    return CT;
    break;
  case Pskel:
    return InBuffGroup (CT->acts[0]);
    break;
  case Dskel:
    return InBuffGroup (CT->acts[0]);
    break;
  default:
    printf("InBuffGroup looking at an unimplemented skeleton\n");
    exit(0);
  }
}


// An auxilliary function we will need. Finds the Give Group unless
// it is the buffer.
callnode_t *GetRealGiveGroup(callnode_t *TG, callnode_t *CT, int N);

// Returns a pointer to the leaf activity group in CT which will perform the 
// Give corresponding to group TG's Nth Take. Returns NULL if the 
// appropriate group is not within CT (which is a valid outcome)
// Implicit assumption: the input buffer is stored by exactly one
// group at the lowest level of CT. 
// Implicit assumption: calling process is a member of TG.
// A returned pointer == TG itself
// implies that TG is the buffer holder and so must Take with a 
// memcopy rather than a real communication
callnode_t *GetGiveGroup(callnode_t *TG, callnode_t *CT, int N) {
  callnode_t *GG;

  if (GG = GetRealGiveGroup(TG, CT, N)) {
    return GG;
  } else {
    // couldn't find an ordinary partner, so it must be the input buffer
    return InBuffGroup(CT);
  }
}


// Algorithm :
//             1. look recursively in TG's subtree 
//             2. look in output fringe of TG's sibling activity at this level, if it
//                has one (ie TGs activity is not on the fringe of CT->skel)
//             3. Give up (recursion unwinds, or ultimately GetGiveGroup finds
//                the overall buffer instead)
callnode_t *GetRealGiveGroup(callnode_t *TG, callnode_t *CT, int N) {
  callnode_t *foundgroup;

  // Case 1 above
  switch (CT->skel) {
  case NOskel:
    return NULL;
    break;
  case Pskel:
    foundgroup = GetRealGiveGroup(TG, CT->acts[FrStack[ Top - (TG->level - CT->level - 1) ]->mycol], N);
    if (foundgroup) return foundgroup;
    break;
  case Dskel:
    foundgroup = GetRealGiveGroup(TG, CT->acts[FrStack[ Top - (TG->level - CT->level - 1) ]->mycol], N);
    if (foundgroup) return foundgroup;
    break;
  default:
    printf("GetGiveGroup looking at an unimplemented skeleton\n");
    exit(0);
  }

  // Case 2 above
  switch (CT->skel) {
  case Pskel:
    if (FrStack[ Top - (TG->level - CT->level - 1) ]->mycol > 0) {
      return FringeGiveGroup(CT->acts[FrStack[ Top - (TG->level - CT->level - 1) ]->mycol - 1], N);
    }
    break;
  case Dskel:
    break;

  default:
    printf("GetGiveGroup looking at an inappropriate or NOskel skeleton\n");
    exit(0);
  }

  // Case 3 above
  return NULL;
}


// --------------------------------------------------------------------------------//
// The same for Take
// --------------------------------------------------------------------------------//

// Find the group which holds the output buffer for CT
callnode_t *OutBuffGroup (callnode_t *CT, int root) {
  switch (CT->skel) {
  case NOskel: 
    return CT;
    break;
  case Pskel:
    return OutBuffGroup (CT->acts[CT->Nacts-1],root);
    break;
  case Dskel:
    return OutBuffGroup (CT->acts[0],root);
    break;
  default:
    printf("OutBuffGroup looking at an unimplemented skeleton\n");
    exit(0);
  }
}


// An auxilliary function we will need. Finds the Take Group unless
// it is the buffer.
callnode_t *GetRealTakeGroup(callnode_t *TG, callnode_t *CT, int N);

// Returns a pointer to the leaf activity group in CT which will perform the 
// Give corresponding to group TG's Nth Take. Returns NULL if the 
// appropriate group is not within CT (which is a valid outcome)
// Implicit assumption: the input buffer is stored by exactly one
// group at the lowest level of CT. 
// Implicit assumption: calling process is a member of TG.
// A returned pointer == TG itself
// implies that TG is the buffer holder and so must Take with a 
// memcopy rather than a real communication

callnode_t *GetTakeGroup(callnode_t *TG, callnode_t *CT, int N) {
  callnode_t *GG;

  if (GG = GetRealTakeGroup(TG, CT, N)) {
    return GG;
  } else {
    // couldn't find an ordinary partner, so it must be the output buffer
    return OutBuffGroup(CT,Top);
  }
}

// Algorithm :
//             1. look recursively in TG's subtree 
//             2. look in output fringe of TG's sibling activity at this level, if it
//                has one (ie TGs activity is not on the fringe of CT->skel)
//             3. Give up (recursion unwinds, or ultimately GetTakeGroup finds
//                the overall buffer instead)

callnode_t *GetRealTakeGroup(callnode_t *TG, callnode_t *CT, int N) {
  callnode_t *foundgroup;

  // Case 1 above
  switch (CT->skel) {
  case NOskel:
    return NULL;
    break;
  case Pskel:
    foundgroup = GetRealTakeGroup(TG, CT->acts[FrStack[ Top - (TG->level - CT->level - 1) ]->mycol], N);
    if (foundgroup) return foundgroup;
    break;
  case Dskel:
    foundgroup = GetRealTakeGroup(TG, CT->acts[FrStack[ Top - (TG->level - CT->level - 1) ]->mycol], N);
    if (foundgroup) return foundgroup;
    break;    
  default:
    printf("GetTakeGroup looking at an unimplemented skeleton\n");
    exit(0);
  }

  // Case 2 above
  switch (CT->skel) {
  case Pskel:
    if (FrStack[ Top - (TG->level - CT->level - 1) ]->mycol < FrStack[Top - (TG->level - CT->level - 1) ]->context.p.ns - 1 ) {
      return FringeTakeGroup(CT->acts[FrStack[ Top - (TG->level - CT->level - 1) ]->mycol + 1], N);
    }
    break;
  case Dskel:
    break;
  default:
    printf("GetTakeGroup looking at an inappropriate or NOskel skeleton\n");
    exit(0);
  }

  // Case 3 above
  return NULL;
}

