int DEBUGP = 0; // set to 1 for debug information

// Called with SPLOCAL data
// only the root of the first stage calls this, taking all the SPLOCAL data down from the skeleton
void recvSPlocalthread (int *junk)
{
  int i, j, ROOT;
  int totalrecvlen;
  char **tempbuff;
  int *lenscount;
  char *cpptr;
  callnode_t *inbuffgroup;
  int *inbuffrootranks;   // ranks of the inbuff group processes as they are in the root comm

  // Finding the root
  ROOT = Top - FrStack[Top]->nestlevel + 1;  

  inbuffgroup = InBuffGroup(FrStack[Top]->mycalltree);
  inbuffrootranks = (int *) malloc (sizeof(int)*inbuffgroup->actsz[0]);
  for (i=0; i<inbuffgroup->actsz[0]; i++)
    inbuffrootranks[i] = inbuffgroup->members[0][i*(inbuffgroup->level+1)];

  lenscount = (int *) malloc (sizeof (int));
  tempbuff   = (char **) malloc (sizeof(char *));

  *lenscount = -1;
  group2group (FrStack[ROOT]->allcomm1, FrStack[ROOT]->parentcommsize,
               FrStack[ROOT]->parentcommrank, RECEIVER,  FrStack[Top]->myrank, // *** careful, rank in rcv comm
               FrStack[ROOT]->allcomm2,  FrStack[ROOT]->parentcommsize, 
               FrStack[ROOT]->allranks,
               FrStack[Top]->stagerootonlycomm,
	       1,  
               inbuffrootranks,
	       MPI_INT, mpiintsize, NULL, 0,
	       (char **) &FrStack[ROOT]->blens,
	       lenscount);
  
  // count and recv the data with g2g
  for (totalrecvlen=0, i=0; i<*lenscount; i++)
    totalrecvlen += FrStack[ROOT]->blens[i];

  *tempbuff = (char *) malloc (totalrecvlen*FrStack[Top]->context.p.itemsz[0]);
  group2group (FrStack[ROOT]->allcomm1, FrStack[ROOT]->parentcommsize,
               FrStack[ROOT]->parentcommrank, RECEIVER,  FrStack[Top]->myrank,// *** careful, rank in rcv comm
               FrStack[ROOT]->allcomm2,  FrStack[ROOT]->parentcommsize,
               FrStack[ROOT]->allranks,
               FrStack[Top]->stagerootonlycomm, 
	       1,  
               inbuffrootranks,
               FrStack[Top]->context.p.ty[0],
	       FrStack[Top]->context.p.itemsz[0], 
               NULL,
	       0,
	       tempbuff,
	       &totalrecvlen);

  // set up the buff itself, by copying for now, although could just set pointers into tempbuff?
  FrStack[ROOT]->buff = (char **) malloc (sizeof(char *) * (*lenscount));
  cpptr = *tempbuff;
  for (i=0; i<*lenscount; i++) {
    (FrStack[ROOT]->buff)[i] = 
       (char *) malloc ( (FrStack[ROOT]->blens)[i] * FrStack[Top]->context.p.itemsz[0]);
    memcpy ((void *) (FrStack[ROOT]->buff)[i], (void *) cpptr,
            (FrStack[ROOT]->blens)[i] * FrStack[Top]->context.p.itemsz[0]);
    cpptr += FrStack[ROOT]->blens[i]*FrStack[Top]->context.p.itemsz[0];
  }

  FrStack[ROOT]->inmul = *lenscount; // set new effective inmul for root
  FrStack[ROOT]->bnext = 0;         // set up "pointer" for Take       

  /*
  if (DEBUGP)
    for (i=0; i<*lenscount;i++) {
      printf("****  blens[%d]=%d \n",i,FrStack[ROOT]->blens[i]);
      for (j=0;j<FrStack[ROOT]->blens[i];j++) {
	printf(" buff %d \n",((int *) FrStack[ROOT]->buff[i])[j]);
      }
    }
  */
}


// Called with SPGLOBAL data
// all processes of the first stage active, taking an SPGLOBAL datum down from the skeleton
void recvSPglobalthread (int *inmul)
{
  int i,j;
  int ROOT; // Frame stack index for the root call of this hierarchy
  int *len;
  char **tmpbuff;
  callnode_t *inbuffgroup;
  int *inbuffrootranks;   // ranks of the inbuff group processes as they are in the root comm

  // Finding the root
  ROOT = Top - FrStack[Top]->nestlevel + 1;  

  inbuffgroup = InBuffGroup(FrStack[Top]->mycalltree);
  inbuffrootranks = (int *) malloc (sizeof(int)*inbuffgroup->actsz[0]);
  for (i=0; i<inbuffgroup->actsz[0]; i++)
    inbuffrootranks[i] = inbuffgroup->members[0][i*(inbuffgroup->level+1)];
  
  len = (int *) malloc (sizeof (int));
  tmpbuff  = (char **) malloc (sizeof(char *) * (*inmul)); 

  for (i=0; i<*inmul; i++) {
    *len = -1;
    group2group (FrStack[ROOT]->allcomm1, FrStack[ROOT]->parentcommsize,
                 FrStack[ROOT]->parentcommrank, RECEIVER,  FrStack[Top]->myrank,
                 FrStack[ROOT]->allcomm2,  FrStack[ROOT]->parentcommsize, 
                 FrStack[ROOT]->allranks,
                 FrStack[Top]->mycomm,
		 FrStack[Top]->mycommsize,
		 inbuffrootranks,
		 FrStack[Top]->inty, //*** chg 16 fev to check //  FrStack[Top]->context.p.ty[0],
		 FrStack[Top]->initemsz,  //*** chg 16 fev to check //FrStack[Top]->context.p.itemsz[0],
                 NULL, 0, tmpbuff, len);
    FrStack[ROOT]->blens[i] = *len;
    FrStack[ROOT]->buff[i] = (char *) malloc (sizeof(int) * (*len));
    memcpy((int *) FrStack[ROOT]->buff[i], (int *) *tmpbuff, sizeof(int) * (*len));
    
    /*
      if (DEBUGP) {
    printf("****  blens[%d]=%d - malloc %d  %d\n",i,FrStack[ROOT]->blens[i], sizeof(int) * (*len), sizeof(tmpbuff));
    for (j=0;j<*len;j++) {
      //((int *) FrStack[Top]->buff[i])[j] = ((int *) tmpbuff[0])[j];
      printf(" buff %d \n",((int *) FrStack[ROOT]->buff[i])[j]);
    }
    }
    */
  }
}



// Called once, immediately after construction of the call tree, to 
// copy the input buffer data from the root group to the appropriate
// (hierarchy specific) leaf group
void BuildInputBuffer (void) {
  int i,totalsendlen;
  int ROOT;
  callnode_t *inbuffgroup;
  int *inbuffrootranks;
  char *sndptr;
  pthread_t recvthread;
  void *threadval;

  // Top is the actual level of nesting, and nestlevel indicates where to go back to find
  //   the buffer. We need to add 1 since the info for the first level is 1   
  ROOT = Top - FrStack[Top]->nestlevel + 1;  

  if (DEBUGP) printf("BuildInputBuffer, rank = %d, col = %d, Top = %d, level = %d, ROOT = %d \n",FrStack[Top]->myrank, FrStack[Top]->mycol, Top, FrStack[Top]->nestlevel, ROOT);

  inbuffgroup = InBuffGroup(FrStack[Top]->mycalltree);
  inbuffrootranks = (int *) malloc (sizeof(int)*inbuffgroup->actsz[0]);
  for (i=0; i<inbuffgroup->actsz[0]; i++)
    inbuffrootranks[i] = inbuffgroup->members[0][i*(inbuffgroup->level+1)];

  //--------------------gathering global spread data-------------------------------//
  if (FrStack[ROOT]->inbufspr == SPGLOBAL) {

    sndptr = (char *) FrStack[ROOT]->in;

    if (inbuffgroup == GetMyGroup()) { // I will hold the assembled input buffer, so make it
      //if (DEBUGP) printf("   BuildInputBuffer: inmul = %d \n", FrStack[ROOT]->inmul);
      FrStack[ROOT]->blens = (int *)  malloc (sizeof(int)*FrStack[ROOT]->inmul);
      FrStack[ROOT]->buff  = (char **)malloc (sizeof(char *)*FrStack[ROOT]->inmul);
      // and spark receiver threads on input buffer processes
      pthread_create(&recvthread, NULL, (void* (*)(void *))&recvSPglobalthread, &FrStack[ROOT]->inmul);
    }

     // Now the "senders" loop calling group2group
    for (i=0; i<FrStack[ROOT]->inmul; i++) {
      //if (DEBUGP) printf("   BuildInputBuffer: inlens[%d] = %d \n", i,  FrStack[ROOT]->inlens[i]);      
      group2group (FrStack[ROOT]->allcomm1, FrStack[ROOT]->parentcommsize,
		   FrStack[ROOT]->parentcommrank, SENDER,  FrStack[ROOT]->parentcommrank,
		   FrStack[ROOT]->allcomm2,  FrStack[ROOT]->parentcommsize, 
		   FrStack[ROOT]->allranks,
		   (MPI_Comm) NULL, inbuffgroup->actsz[0], inbuffrootranks,
		   FrStack[ROOT]->inty, FrStack[ROOT]->initemsz,
		   sndptr, FrStack[ROOT]->inlens[i], NULL, NULL);
      sndptr = sndptr + (FrStack[ROOT]->initemsz * FrStack[ROOT]->inlens[i]);
    }
    FrStack[ROOT]->bnext = 0; // set up "pointer" for Take

    if (inbuffgroup == GetMyGroup())
      pthread_join(recvthread, &threadval);
  }


  //--------------------gathering local spread data-------------------------------//
  else {

    // spark receiver thread on in buffer group root
    if (inbuffgroup == GetMyGroup() && FrStack[Top]->myrank == 0) {
      pthread_create(&recvthread, NULL, (void* (*)(void *))&recvSPlocalthread, NULL);
    }
    // send the lengths with a g2g
    group2group (FrStack[ROOT]->allcomm1, FrStack[ROOT]->parentcommsize,
                 FrStack[ROOT]->parentcommrank, SENDER,  FrStack[ROOT]->parentcommrank,
                 FrStack[ROOT]->allcomm2,  FrStack[ROOT]->parentcommsize, FrStack[ROOT]->allranks,
                 (MPI_Comm) NULL, 1, inbuffrootranks,  // actually only the first will be used
                 MPI_INT, mpiintsize,
		 (char *) FrStack[ROOT]->inlens, FrStack[ROOT]->inmul, NULL, NULL);
    
    // count and send the data with g2g
    for (totalsendlen=0, i=0; i<FrStack[ROOT]->inmul; i++)
      totalsendlen += FrStack[ROOT]->inlens[i];    

    group2group (FrStack[ROOT]->allcomm1, FrStack[ROOT]->parentcommsize,
                 FrStack[ROOT]->parentcommrank, SENDER,  FrStack[ROOT]->parentcommrank,
                 FrStack[ROOT]->allcomm2,  FrStack[ROOT]->parentcommsize, 
                 FrStack[ROOT]->allranks,
                 (MPI_Comm) NULL, 1, inbuffrootranks,
                 FrStack[ROOT]->inty, FrStack[ROOT]->initemsz, 
                 FrStack[ROOT]->in, totalsendlen, NULL, NULL);
     
    if (inbuffgroup == GetMyGroup() && FrStack[Top]->myrank == 0)
      pthread_join(recvthread, &threadval);
    
  }
  
  if (DEBUGP) printf("End BuildInputBuffer - %d %d \n", FrStack[ROOT]->mycol, FrStack[ROOT]->myrank);

}


// all processes of the final stage active, sending an SPGLOBAL result back up to the skeleton
// Called by ReturnOutputBuffer
void pipesendthread (int *outmul) {
  int i;
  callnode_t *outbuffgroup;
  int *outbuffrootranks;
  int ROOT;

  ROOT = Top - FrStack[Top]->nestlevel + 1;  

  //if (DEBUGP) printf("Pipesendthread, Top %d, ROOT %d\n", Top, ROOT);

  // Looking at the group holding the output buffer (last stage group)
  outbuffgroup = OutBuffGroup(FrStack[Top]->mycalltree,ROOT);
  outbuffrootranks = (int *) malloc (sizeof(int) * outbuffgroup->actsz[0]);
  for (i=0; i<outbuffgroup->actsz[0]; i++)
    outbuffrootranks[i] = outbuffgroup->members[0][i*(outbuffgroup->level+1)]; 

  for (i=0; i<*outmul; i++) {
    group2group (FrStack[ROOT]->allcomm1, FrStack[ROOT]->parentcommsize,
		 FrStack[ROOT]->parentcommrank, SENDER,  FrStack[Top]->myrank,
		 FrStack[Top]->mycomm,  FrStack[Top]->mycommsize,  
		 outbuffrootranks,
		 FrStack[ROOT]->allcomm2, FrStack[ROOT]->parentcommsize, 
		 FrStack[ROOT]->allranks,
		 FrStack[ROOT]->context.p.ty[FrStack[ROOT]->context.p.ns], 
		 FrStack[ROOT]->context.p.itemsz[FrStack[ROOT]->context.p.ns],
		 FrStack[ROOT]->poutbuff[i], FrStack[ROOT]->poutblens[i], NULL, NULL);
  }
}


// Called once, immediately after construction of the call tree,
// to make the temporary output buffer in the appropriate
// (hierarchy specific) leaf group
// Need to handle unknown number of outputs                 
void BuildOutputBuffer (void) {
  int i, ROOT;
  callnode_t *outbuffgroup;
  int *outbuffrootranks;    // ranks of the outbuff group processes as they are in the root comm

  // Top is the actual level of nesting, and nestlevel indicates where to go back to find
  // the buffer. We need to add 1 since the info for the first level is 1      
  ROOT = Top - FrStack[Top]->nestlevel + 1;  

  // Looking at the group holding the output buffer (last stage group)
  outbuffgroup = OutBuffGroup(FrStack[Top]->mycalltree,ROOT);
  outbuffrootranks = (int *) malloc (sizeof(int) * outbuffgroup->actsz[0]);
  for (i=0; i<outbuffgroup->actsz[0]; i++) {
    outbuffrootranks[i] = outbuffgroup->members[0][i*(outbuffgroup->level+1)];
  }
  
  // Processes of outbuffrootranks: create output buffers
  if (outbuffgroup == GetMyGroup()) {
    // Global spread: data distributed between all processes working on the last stage
    if (FrStack[ROOT]->outbufspr == SPGLOBAL) {
      FrStack[ROOT]->poutblens = (int *)   malloc (sizeof(int)*MAXOUTMUL);
      FrStack[ROOT]->poutbuff  = (char **) malloc (sizeof(char *)*MAXOUTMUL);
      FrStack[ROOT]->poutbnext = 0;
    }
    else { 
      // Local spread: everything will be on the process root of the outbuffgroup.
      if (FrStack[ROOT]->parentcommrank == outbuffrootranks[0]) {
        FrStack[ROOT]->poutblens = (int *)  malloc (sizeof(int)*MAXOUTMUL);
        FrStack[ROOT]->poutbuff  = (char **)malloc (sizeof(char *)*MAXOUTMUL);
        FrStack[ROOT]->poutbnext = 0;
      }
    }
  }

  return;
}



// Called once, immediately after termination of the leaf
// activities in a hierarchy, to copy the output
// buffer data from the leaves to the hierarchy root group
void ReturnOutputBuffer (void) {

  int i,temp, ROOT;
  callnode_t *outbuffgroup;
  int *outbuffrootranks;    // ranks of the outbuff group processes as they are in the root comm
  int outbuffspaceleft;     // during output, how much space is left in the out buffer    
  int outbufflensspaceleft; // during output, how much space is left in the outlens buffer
  char *rcvptr;
  pthread_t sendthread;
  void *threadval;      
  int *tempptr;
  char **rcvptrptr;

  // Top is the actual level of nesting, and nestlevel indicates where to go back to find
  // the buffer. We need to add 1 since the info for the first level is 1      
  ROOT = Top - FrStack[Top]->nestlevel + 1;  
  
  if (DEBUGP)
    printf("*** Returning output buffer... Top %d ROOT %d stage %d rank %d\n", Top, ROOT, FrStack[ROOT]->mycol,FrStack[ROOT]->myrank);
  
  // Looking at the group holding the output buffer (last stage group)
  outbuffgroup = OutBuffGroup(FrStack[Top]->mycalltree,ROOT);
  outbuffrootranks = (int *) malloc (sizeof(int) * outbuffgroup->actsz[0]);
  for (i=0; i<outbuffgroup->actsz[0]; i++)
    outbuffrootranks[i] = outbuffgroup->members[0][i*(outbuffgroup->level+1)];


  // Distribute output from buffer at final stage  

  // -------------------------- ReturnOutputBuffer for SPGLOBAL --------------------------//
  if (FrStack[ROOT]->outbufspr == SPGLOBAL) {
    outbuffspaceleft     = FrStack[ROOT]->outbuffsz;
    outbufflensspaceleft = FrStack[ROOT]->outlenP[0][1]; 

    // broadcast actual multiplicity of output
    MPI_Bcast(&FrStack[ROOT]->poutbnext, 1, MPI_INT,
	      outbuffrootranks[0], // root of the last stage
	      FrStack[ROOT]->parentcomm);
    
    FrStack[ROOT]->blens = (int *)  malloc (sizeof(int)* FrStack[ROOT]->poutbnext);
    rcvptr = (char *) FrStack[ROOT]->out;

    // Now the "receivers" loop calling group2group
    // spark sender threads on stage ns-1 processes
    
    if (outbuffgroup == GetMyGroup())
      pthread_create(&sendthread, NULL, (void* (*)(void *))&pipesendthread, &(FrStack[ROOT]->poutbnext));
    
    tempptr = (int *) malloc(sizeof(int));
    rcvptrptr = (char **) malloc(sizeof(char *));
    for  (i=0; i<FrStack[ROOT]->poutbnext; i++) {
      *rcvptrptr = rcvptr;
      *tempptr = outbuffspaceleft;

      group2group (FrStack[ROOT]->allcomm1, FrStack[ROOT]->parentcommsize,
		   FrStack[ROOT]->parentcommrank, RECEIVER,  FrStack[ROOT]->parentcommrank,
		   FrStack[ROOT]->mycomm,  FrStack[ROOT]->mycommsize,  
		   outbuffrootranks, // array with the last stage processes
		   FrStack[ROOT]->allcomm2, FrStack[ROOT]->parentcommsize, 
		   FrStack[ROOT]->allranks,
		   FrStack[ROOT]->context.p.ty[FrStack[ROOT]->context.p.ns],
		   FrStack[ROOT]->context.p.itemsz[FrStack[ROOT]->context.p.ns],
		   NULL, 0, rcvptrptr, tempptr);
      outbuffspaceleft -= *tempptr;

      if (i<outbufflensspaceleft) {
	FrStack[ROOT]->outlenP[0][i] = *tempptr;
      } else {
        fprintf(stderr, "No more space for pipeline outbufflens\n");
	exit(0);
      } 
      rcvptr += ((*tempptr)*FrStack[ROOT]->context.p.itemsz[FrStack[ROOT]->context.p.ns]);
    } 
 
    if (outbuffgroup == GetMyGroup()) 
      pthread_join(sendthread, &threadval);
    
    **(FrStack[ROOT]->outmulP) = FrStack[ROOT]->poutbnext;
  } 

  // ----------------- ReturnOutputBuffer for SPLOCAL -------------------------//
  else {    // distribute local spread results
    // It would be "neat" to do this with g2g as well, but it just doesn't fit because of the 
    // possibility of varying item length in pipelinev. If the data length is uniform, eg as 
    // in pipeline, then it would fit.                                                       
    
    Local_Mult_Scatter (FrStack[ROOT]->parentcomm, FrStack[ROOT]->parentcommsize,
			FrStack[ROOT]->parentcommrank,
			outbuffrootranks[0], FrStack[Top]->context.p.ty[FrStack[Top]->context.p.ns],
			FrStack[Top]->context.p.itemsz[FrStack[Top]->context.p.ns],
			(void **) FrStack[ROOT]->poutbuff , FrStack[ROOT]->poutbnext, 
			FrStack[ROOT]->poutblens, (char *) FrStack[ROOT]->out,
			FrStack[ROOT]->outmulP[0], FrStack[ROOT]->outlenP[0]);
  }
  
  if (DEBUGP) printf("End ReturnOutBuff %d \n",FrStack[0]->myrank);
  return;
}



void PipeGive (void *out[], int len[]) {

  int i, j;
  int  *inlens, *indisp;
  char *givebuff;
  char *sndptr;
  int gtysz, thissz, totallen;
  int mylevel, mysize, *mymembers, *gmyranks;
  int partnerlevel, partnersize, *partnermembers, *gpartnerranks;
  MPI_Status stat;
  int junk, ROOT;
  callnode_t *mynode, *partnernode, *partnerfather;
  
  ROOT = Top - FrStack[Top]->nestlevel + 1;  
  
  if (DEBUGP) printf("PIPEGIVE - Top %d, rank %d, col %d - ROOT %d, rank %d, col %d\n", Top, FrStack[Top]->myrank, FrStack[Top]->mycol, ROOT, FrStack[ROOT]->myrank, FrStack[ROOT]->mycol);

  if (FrStack[Top]->mycalltree == NULL) {
    // This must be the first interaction, so build the call tree
    BuildCallTree();
    if (DEBUGP) if ((FrStack[ROOT]->myrank == 0)&&(FrStack[ROOT]->mycol == 0)) DisplayCallTree(FrStack[Top]->mycalltree);
    
    // and then copy down the input buffer from the root of this call hierarchy,
    BuildInputBuffer();    
    // and allocate space for the output buffer.
    BuildOutputBuffer();
  }

  // Now get on with the Take
  mynode = GetMyGroup();
  mylevel = mynode->level;
  mysize  = mynode->actsz[0];
  mymembers = mynode->members[0];

  gmyranks  = (int *) malloc(sizeof(int)*mysize);
  for (i=0; i<mysize; i++) 
    gmyranks[i]  = mymembers[i*(mylevel+1)];

  if (DEBUGP) {
  printf("Group:  || ");
  for (i=0; i<mysize; i++) {
    for (j=0; j<mylevel+1; j++) {
      printf("%d ",mymembers[i*(mylevel+1) + j]);
    }
    printf("|| ");
  }
  printf(" giving to Take Group:  || ");
  }

  partnernode = GetTakeGroup(mynode, FrStack[Top]->mycalltree, FrStack[Top]->nbgive);
  FrStack[Top]->nbgive += 1;
  partnerlevel = partnernode->level;
  partnersize  = partnernode->actsz[0];
  partnermembers = partnernode->members[0]; 

  gpartnerranks = (int *) malloc(sizeof(int)*partnersize);
  for (i=0; i<partnersize; i++)
    gpartnerranks[i] = partnermembers[i*(partnerlevel+1)];


  if (DEBUGP) {
  if (partnermembers) {
    for (i=0; i<partnersize; i++) {
      for (j=0; j<partnerlevel+1; j++) {
        printf("%d ",partnermembers[i*(partnerlevel+1) + j]);
      }
      printf("|| ");
    }
    printf("\n");
  } else {
    printf("unknown group ||\n");
  }
  }



  //------------------------------------------ Give SPGLOBAL --------------------------------------/
  if (FrStack[Top]->context.p.spr[FrStack[Top]->mycol+1] == SPGLOBAL) {
    
    if (len[0] == SHUTDOWNLEN) {
      // pass on the shut down message
      if (DEBUGP) printf("Give-SPGLOBAL: shutdown message\n");
      if (FrStack[Top]->myrank == 0) {
        for (i=0; i<partnersize; i++) {
	  MPI_Send(NULL, 0, MPI_INT, gpartnerranks[i],
		   SHUTDOWNTAG, FrStack[ROOT]->parentcomm);
        }
      }
      
      // If the partner is a Deal, needs to go on passing the shutdowntag through
      // standard Give. Need to add another tag to confirm end of shutdown
      partnerfather = ((FrStack[Top]->mycalltree)->acts[FrStack[ Top - (mynode->level - (FrStack[Top]->mycalltree)->level - 1) ]->mycol + 1]);

      if ((FrStack[Top]->context.p.shutdown != ENDSHUTDOWN) && (partnerfather->skel == Dskel)) {
	FrStack[Top]->context.p.shutdown = ENDSHUTDOWN;
	for (i=0; i<partnerfather->Nacts-1; i++)
	  PipeGive(out,len);
	ReturnOutputBuffer();
      }	
      // If the partner father is not a deal, there's never the endshutdown tag
      else if (FrStack[Top]->context.p.shutdown != ENDSHUTDOWN)
	ReturnOutputBuffer();
      
      return;
    }
    if (mynode == partnernode) {
      // last stage so just buffer data for later collection
      if (DEBUGP) printf("Give-SPGLOBAL: last stage\n");
      FrStack[ROOT]->poutblens[FrStack[ROOT]->poutbnext] = len[0];
      FrStack[ROOT]->poutbuff[FrStack[ROOT]->poutbnext] =
        (char *) malloc (len[0]*FrStack[Top]->context.p.itemsz[FrStack[Top]->mycol+1]);
      memcpy((void *) FrStack[ROOT]->poutbuff[FrStack[ROOT]->poutbnext],
	     (char *) out[0], len[0]*FrStack[Top]->context.p.itemsz[FrStack[Top]->mycol+1]);
      FrStack[ROOT]->poutbnext++;
    }
    else { 
      // other stage, so interact with next stage
      if (DEBUGP) printf("Give-SPGLOBAL: interact with next stage\n");
      
      // first indicate that this is a real transfer
      if (FrStack[Top]->myrank == 0) {
        for (i=0; i<partnersize; i++) {
	  MPI_Send(NULL, 0, MPI_INT,
		   partnermembers[i*(partnerlevel+1)], 
		   NOSHUTDOWNTAG, FrStack[ROOT]->parentcomm);
	}
      }

      // and then send the stuff    
      group2group (FrStack[ROOT]->parentcomm, FrStack[ROOT]->parentcommsize,
                   FrStack[ROOT]->parentcommrank, SENDER,  FrStack[Top]->myrank,
                   FrStack[Top]->mycomm,  FrStack[Top]->mycommsize, gmyranks,
		   MPI_COMM_NULL, partnersize, gpartnerranks, 
                   FrStack[Top]->context.p.ty[FrStack[Top]->mycol+1],
                   FrStack[Top]->context.p.itemsz[FrStack[Top]->mycol+1],
                   (char *) out[0], len[0], NULL, NULL);
    }
  }
  

  //------------------------------------------ Give SPLOCAL --------------------------------------/
  else if (FrStack[Top]->context.p.spr[FrStack[Top]->mycol+1] == SPLOCAL) {
    
    if (FrStack[Top]->myrank == 0) {  // I am the root of the stage
      if (len[0] == SHUTDOWNLEN) {
	if (DEBUGP) printf("Give-SPLOCAL: the root of the stage pass on the shutdown message\n");
        // pass on the shutdown message
        MPI_Send(NULL, 0,  FrStack[Top]->context.p.ty[FrStack[Top]->mycol+1],
                 partnermembers[0], SHUTDOWNTAG, FrStack[ROOT]->parentcomm);
	// and build the output buffer
	ReturnOutputBuffer();
        return;
      }

      if (mynode == partnernode) {
	if (DEBUGP) printf("Give-SPLOCAL: last stage write into buffer\n");
        // last stage so buffer it until later
        FrStack[ROOT]->poutblens[FrStack[ROOT]->poutbnext] = len[0];
        FrStack[ROOT]->poutbuff[FrStack[ROOT]->poutbnext] =
          (char *) malloc (len[0]*FrStack[Top]->context.p.itemsz[FrStack[Top]->mycol+1]);
        memcpy((void *) FrStack[ROOT]->poutbuff[FrStack[ROOT]->poutbnext],
               (char *) out[0], len[0]*FrStack[Top]->context.p.itemsz[FrStack[Top]->mycol+1]);
        FrStack[ROOT]->poutbnext++;
      } else {
        // Send the item from the root of this stage to the root of the next, via the parent comm
	if (DEBUGP) printf("Give-SPLOCAL gives %d data to %d\n", len[0],partnermembers[0]);
        MPI_Send((char *) out[0], len[0],  FrStack[Top]->context.p.ty[FrStack[Top]->mycol+1],
                 partnermembers[0], NOSHUTDOWNTAG,  FrStack[ROOT]->parentcomm);
      }
    }
    else {
      // I'm not the root but still need to perform the ReturnOutputBuffer if needed
      if (len[0] == SHUTDOWNLEN) {
	ReturnOutputBuffer();
	return;
      }
    }
  }
  return;
}




void **PipeTake (int **len) {
  int i, j;
  int mylevel, mysize, *mymembers, *gmyranks;
  int partnerlevel, partnersize, *partnermembers, *gpartnerranks;
  int junk, ROOT;
  MPI_Status stat;
  void **takeptr2;
  callnode_t *mynode, *partnernode;
  
  /* Top is the actual level of nesting, and nestlevel indicates where to go back to find
     the buffer. We need to add 1 since the info for the first level is 1                   */
  ROOT = Top - FrStack[Top]->nestlevel + 1;  
  
  if (DEBUGP) printf("PIPETAKE - Top %d, rank %d, col %d - ROOT %d, rank %d, col %d\n", Top, FrStack[Top]->myrank, FrStack[Top]->mycol, ROOT, FrStack[ROOT]->myrank, FrStack[ROOT]->mycol);

  if (FrStack[Top]->mycalltree == NULL) {
    // This must be the first interaction, so build the call tree
    BuildCallTree();
    if (DEBUGP) if ((FrStack[ROOT]->myrank == 0)&&(FrStack[ROOT]->mycol == 0)) DisplayCallTree(FrStack[Top]->mycalltree); 

    // and then copy down the input buffer from the root of this call hierarchy,
    BuildInputBuffer();    
    // and allocate space for the output buffer.
    BuildOutputBuffer();
  }

  // Now get on with the Give
  mynode = GetMyGroup();
  mylevel = mynode->level;
  mysize  = mynode->actsz[0];
  mymembers = mynode->members[0];

  gmyranks  = (int *) malloc(sizeof(int)*mysize);
  for (i=0; i<mysize; i++) 
    gmyranks[i]  = mymembers[i*(mylevel+1)];

  if (DEBUGP) {
  printf("Group:  || ");
  for (i=0; i<mysize; i++) {
    for (j=0; j<mylevel+1; j++) {
      printf("%d ",mymembers[i*(mylevel+1) + j]);
    }
    printf("|| ");
  }
  printf(" taking from Give Group:  || ");
  }

  partnernode = GetGiveGroup(mynode, FrStack[Top]->mycalltree, FrStack[Top]->nbtake);
  FrStack[Top]->nbtake += 1;
  partnerlevel = partnernode->level;
  partnersize  = partnernode->actsz[0];
  partnermembers = partnernode->members[0]; 

  gpartnerranks = (int *) malloc(sizeof(int)*partnersize);
  for (i=0; i<partnersize; i++)
    gpartnerranks[i] = partnermembers[i*(partnerlevel+1)];

  if (DEBUGP) {
  if (partnermembers) {
    for (i=0; i<partnersize; i++) {
      for (j=0; j<partnerlevel+1; j++) {
        printf("%d ",partnermembers[i*(partnerlevel+1) + j]);
      }
      printf("|| ");
    }
    printf("\n");
  } else {
    printf("unknown group ||\n");
  }
  }
 
  // Take from the give group or from the input buffer, depending on the case
  
  //--------------------- Take SPGLOBAL ---------------------------------------//
  if (FrStack[Top]->context.p.spr[FrStack[Top]->mycol] == SPGLOBAL) {
    if (FrStack[Top]->context.p.shutdown) {
      printf("Trying to Take after shutdown....\n");
      exit(0);
    }
    if (mynode == partnernode) {
      // get data from my buffer or return FALSE if buffer is empty and set shutdown = TRUE
      if (FrStack[ROOT]->bnext == FrStack[ROOT]->inmul) {
	if (DEBUGP) printf("Take-SPGLOBAL: first stage find empty buffer \n");
	FrStack[Top]->context.p.shutdown = TRUE;
	// Special case where it's also the last stage: should build the output buffer
	if (mynode == OutBuffGroup(FrStack[Top]->mycalltree,ROOT))
	  ReturnOutputBuffer();
	return NULL;
      }
      // non-empty buffer
      if (DEBUGP) printf("Take-SPGLOBAL: first stage find non empty buffer \n");
      // set the length
      *len = (int *) malloc (sizeof(int));
      **len = FrStack[ROOT]->blens[FrStack[ROOT]->bnext]; 
      // set the data
      takeptr2 = (void **) malloc (sizeof(void *));
      *takeptr2 = (void *) malloc (FrStack[Top]->context.p.itemsz[FrStack[Top]->mycol] * (**len));
      memcpy(*takeptr2, ((int *) FrStack[ROOT]->buff[FrStack[ROOT]->bnext]),
	     FrStack[Top]->context.p.itemsz[FrStack[Top]->mycol]* (**len));
      FrStack[ROOT]->bnext++;
      return takeptr2;
    }
    else {
      // other stage so interact with previous stage
      MPI_Recv(&junk, 1, MPI_INT, gpartnerranks[0], MPI_ANY_TAG, FrStack[ROOT]->parentcomm, &stat);
      if (stat.MPI_TAG==SHUTDOWNTAG) {
	if (DEBUGP) printf("Take-SPGLOBAL: process %d find shutdown tag\n", FrStack[ROOT-1]->myrank);	
	FrStack[Top]->context.p.shutdown = TRUE;
	// Build the output buffer if it is the very last stage
	if (mynode == OutBuffGroup(FrStack[Top]->mycalltree,ROOT))
	  ReturnOutputBuffer();
	return NULL;
      }
      // get the stuff
      if (DEBUGP) printf("Take-SPGLOBAL: interact with previous stage\n");
      takeptr2 = (void **) malloc (sizeof(void *));
      *len = (int *) malloc (sizeof(int));
      i=-1;
      **len = i;
      
      group2group (FrStack[ROOT]->parentcomm, FrStack[ROOT]->parentcommsize,
		   FrStack[ROOT]->parentcommrank, RECEIVER,  FrStack[Top]->myrank,
		   MPI_COMM_NULL,  partnersize, gpartnerranks,
		   FrStack[Top]->mycomm, FrStack[Top]->mycommsize, gmyranks,
		   FrStack[Top]->context.p.ty[FrStack[Top]->mycol],
		   FrStack[Top]->context.p.itemsz[FrStack[Top]->mycol],
		   NULL, 0, (char **) takeptr2, *len);

      return takeptr2;
    }
  }

 
  //--------------------- Take SPLOCAL ---------------------------------------//
  else if (FrStack[Top]->context.p.spr[FrStack[Top]->mycol] == SPLOCAL) {    
    
    if (FrStack[Top]->myrank == 0) {
      // I'm the root of the stage
      if (mynode == partnernode) {
	// First Stage, take input from the buffer or shutdown if it's empty
	if (FrStack[ROOT]->bnext == FrStack[ROOT]->inmul) {
	  // Buffer empty, broadcast shutdown tag to other members
	  if (DEBUGP) printf("Take-SPLOCAL: root of first stage find empty buffer \n");
	  junk = SHUTDOWNTAG;
	  MPI_Bcast(&junk, 1, MPI_INT, 0, FrStack[Top]->mycomm);
	  FrStack[Top]->context.p.shutdown = TRUE;
	  // Special case where it's also the last stage: should build the output buffer
	  if (mynode == OutBuffGroup(FrStack[Top]->mycalltree,ROOT))
	    ReturnOutputBuffer();
	  return NULL;
	}
	// Non-empty buffer
	if (DEBUGP) printf("Take-SPLOCAL: root of first stage find non empty buffer \n");
	*len = (int *) malloc (sizeof(int));
	**len = FrStack[ROOT]->blens[FrStack[ROOT]->bnext]; // set the length
	takeptr2 = (void **) malloc (sizeof(void *));    
	*takeptr2 = (void *) malloc (FrStack[Top]->context.p.itemsz[0] * (**len)); // set the data
	memcpy(*takeptr2, (int *) FrStack[ROOT]->buff[FrStack[ROOT]->bnext],
	       FrStack[Top]->context.p.itemsz[0]* (**len));
	FrStack[ROOT]->bnext++;
	// Broadcast noshutdown tag to other stage members
	junk = NOSHUTDOWNTAG;
	MPI_Bcast(&junk, 1, MPI_INT, 0, FrStack[Top]->mycomm);
	return takeptr2;
      }
      else {// Root of another stage
	// Receive data from the preceding root
	if (DEBUGP) printf("Take-SPLOCAL: root of stage %d interact with previous stage \n", FrStack[Top]->mycol + 1);

	*len = (int *) malloc (sizeof(int));
	
	MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, FrStack[ROOT]->parentcomm, &stat);
	
	if (stat.MPI_TAG==SHUTDOWNTAG) {
	  if (DEBUGP) printf("... and find shutdowntag\n");
	  // receive the "message"
	  takeptr2 = (void **) malloc (sizeof(void *));
	  *takeptr2 = (void *) malloc (FrStack[Top]->context.p.itemsz[FrStack[Top]->mycol]);  // dummy
	  MPI_Recv(takeptr2, 1, FrStack[Top]->context.p.ty[FrStack[Top]->mycol],
		   MPI_ANY_SOURCE, MPI_ANY_TAG, FrStack[ROOT]->parentcomm, &stat);
	  // broadcast shutdown tag to other stage members
	  junk = SHUTDOWNTAG;
	  MPI_Bcast (&junk, 1, MPI_INT, 0, FrStack[Top]->mycomm);
	  FrStack[Top]->context.p.shutdown = TRUE;
	  // Build the output buffer if it is the very last stage
	  if (mynode == OutBuffGroup(FrStack[Top]->mycalltree,ROOT))
	    ReturnOutputBuffer();
	  return NULL;
	} else {
	  if (DEBUGP) printf("... and find no shutdowntag  \n" );
	  MPI_Get_count(&stat, FrStack[Top]->context.p.ty[FrStack[Top]->mycol], *len);
	  takeptr2 = (void **) malloc (sizeof(void *));
	  *takeptr2 = (void *) malloc (FrStack[Top]->context.p.itemsz[FrStack[Top]->mycol] * (*len[0]));
	  MPI_Recv(*takeptr2, **len, FrStack[Top]->context.p.ty[FrStack[Top]->mycol],
		   MPI_ANY_SOURCE, MPI_ANY_TAG, FrStack[ROOT]->parentcomm, &stat);
	  junk = NOSHUTDOWNTAG;
	  MPI_Bcast (&junk, 1, MPI_INT, 0, FrStack[Top]->mycomm);

	  return takeptr2;
	}	
      }
    }
    else {
      // I'm not the root
      if (DEBUGP) printf("Take-SPLOCAL: i'm not the root\n");

      MPI_Bcast (&junk, 1, MPI_INT, 0, FrStack[Top]->mycomm);
      if (junk==SHUTDOWNTAG) {
	FrStack[Top]->context.p.shutdown = TRUE;
	// output buffer if I am the really last stage
	if (mynode == OutBuffGroup(FrStack[Top]->mycalltree,ROOT))
	  ReturnOutputBuffer();
	return NULL;
      } else {
	takeptr2 = (void **) malloc (sizeof(void *));    
	*takeptr2 = (void *) malloc (sizeof(void *));
	return takeptr2;
      }      
    }
  }  
  return NULL;
}
